/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TwitterNotifier.
 *
 * The Initial Developer of the Original Code is
 * Kazuho Okui.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla code.
 *
 * The Initial Developer of the Original Code is
 * Simon Bünzli <zeniko@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2006-2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Utilities for JavaScript code to handle JSON content.
 * See http://www.json.org/ for comprehensive information about JSON.
 *
 * Import this module through
 *
 * Components.utils.import("resource://gre/modules/JSON.jsm");
 *
 * Usage:
 *
 * var newJSONString = JSON.toString( GIVEN_JAVASCRIPT_OBJECT );
 * var newJavaScriptObject = JSON.fromString( GIVEN_JSON_STRING );
 *
 * Note: For your own safety, Objects/Arrays returned by
 *       JSON.fromString aren't instanceof Object/Array.
 */

var EXPORTED_SYMBOLS = ["JSON"];

// The following code is a loose adaption of Douglas Crockford's code
// from http://www.json.org/json.js (public domain'd)

// Notable differences:
// * Unserializable values such as |undefined| or functions aren't
//   silently dropped but always lead to a TypeError.
// * An optional key blacklist has been added to JSON.toString

var JSON = {
  /**
   * Converts a JavaScript object into a JSON string.
   *
   * @param aJSObject is the object to be converted
   * @param aKeysToDrop is an optional array of keys which will be
   *                    ignored in all objects during the serialization
   * @return the object's JSON representation
   *
   * Note: aJSObject MUST not contain cyclic references.
   */
  toString: function JSON_toString(aJSObject, aKeysToDrop) {
    // we use a single string builder for efficiency reasons
    var pieces = [];
    
    // this recursive function walks through all objects and appends their
    // JSON representation (in one or several pieces) to the string builder
    function append_piece(aObj) {
      if (typeof aObj == "string") {
        aObj = aObj.replace(/[\\"\x00-\x1F\u0080-\uFFFF]/g, function($0) {
          // use the special escape notation if one exists, otherwise
          // produce a general unicode escape sequence
          switch ($0) {
          case "\b": return "\\b";
          case "\t": return "\\t";
          case "\n": return "\\n";
          case "\f": return "\\f";
          case "\r": return "\\r";
          case '"':  return '\\"';
          case "\\": return "\\\\";
          }
          return "\\u" + ("0000" + $0.charCodeAt(0).toString(16)).slice(-4);
        });
        pieces.push('"' + aObj + '"')
      }
      else if (typeof aObj == "boolean") {
        pieces.push(aObj ? "true" : "false");
      }
      else if (typeof aObj == "number" && isFinite(aObj)) {
        // there is no representation for infinite numbers or for NaN!
        pieces.push(aObj.toString());
      }
      else if (aObj === null) {
        pieces.push("null");
      }
      // if it looks like an array, treat it as such - this is required
      // for all arrays from either outside this module or a sandbox
      else if (aObj instanceof Array ||
               typeof aObj == "object" && "length" in aObj &&
               (aObj.length === 0 || aObj[aObj.length - 1] !== undefined)) {
        pieces.push("[");
        for (var i = 0; i < aObj.length; i++) {
          arguments.callee(aObj[i]);
          pieces.push(",");
        }
        if (aObj.length > 0)
          pieces.pop(); // drop the trailing colon
        pieces.push("]");
      }
      else if (typeof aObj == "object") {
        pieces.push("{");
        for (var key in aObj) {
          // allow callers to pass objects containing private data which
          // they don't want the JSON string to contain (so they don't
          // have to manually pre-process the object)
          if (aKeysToDrop && aKeysToDrop.indexOf(key) != -1)
            continue;
          
          arguments.callee(key.toString());
          pieces.push(":");
          arguments.callee(aObj[key]);
          pieces.push(",");
        }
        if (pieces[pieces.length - 1] == ",")
          pieces.pop(); // drop the trailing colon
        pieces.push("}");
      }
      else {
        throw new TypeError("No JSON representation for this object!");
      }
    }
    append_piece(aJSObject);
    
    return pieces.join("");
  },

  /**
   * Converts a JSON string into a JavaScript object.
   *
   * @param aJSONString is the string to be converted
   * @return a JavaScript object for the given JSON representation
   */
  fromString: function JSON_fromString(aJSONString) {
    if (!this.isMostlyHarmless(aJSONString))
      throw new SyntaxError("No valid JSON string!");
    
    var s = new Components.utils.Sandbox("about:blank");
    return Components.utils.evalInSandbox("(" + aJSONString + ")", s);
  },

  /**
   * Checks whether the given string contains potentially harmful
   * content which might be executed during its evaluation
   * (no parser, thus not 100% safe! Best to use a Sandbox for evaluation)
   *
   * @param aString is the string to be tested
   * @return a boolean
   */
  isMostlyHarmless: function JSON_isMostlyHarmless(aString) {
    const maybeHarmful = /[^,:{}\[\]0-9.\-+Eaeflnr-u \n\r\t]/;
    const jsonStrings = /"(\\.|[^"\\\n\r])*"/g;
    
    return !maybeHarmful.test(aString.replace(jsonStrings, ""));
  }
};

const CLASS_ID = Components.ID("3c4bfeef-c936-0d98-8fb9-b9b28e4f28ed");
const CLASS_NAME = "Twitter Notifier";
const CONTRACT_ID = "@naan.studio/twitternotifier;1";
const NETWORK_TIMEOUT_TIME = 120;
const MAX_STORED_MESSAGES = 40;

const TWITTER_API_URL = "https://twitter.com/";
const APP_NAME = "TwitterFox";

const STATE_ACTIVE  = 0;
const STATE_IDLE    = 1;
const STATE_SUSPEND = 2;

const APPEND_TIMELINE = 0;
const APPEND_MESSAGES = 1;
const APPEND_IGNORE   = 2;

var keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

function toUTF8Octets(string) {
  return unescape(encodeURIComponent(string));
}

function btoa(input) {
  var output = "";
  var chr1, chr2, chr3;
  var enc1, enc2, enc3, enc4;
  var i = 0;

  do {
    chr1 = input.charCodeAt(i++);
    chr2 = input.charCodeAt(i++);
    chr3 = input.charCodeAt(i++);

    enc1 = chr1 >> 2;
    enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
    enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
    enc4 = chr3 & 63;

    if (isNaN(chr2)) {
      enc3 = enc4 = 64;
    } else if (isNaN(chr3)) {
      enc4 = 64;
    }

    output = output + keyStr.charAt(enc1) + keyStr.charAt(enc2) +
      keyStr.charAt(enc3) + keyStr.charAt(enc4);
  } while (i < input.length);

  return output;
}

function HttpRequest() {
  this.responseText = "";
  this.status = 0;

  var observer = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  observer.addObserver(this, "http-on-modify-request", false);
  observer.addObserver(this, "http-on-examine-response", false);
}

HttpRequest.prototype = {

  httpChannel: function() {
    return this.channel.QueryInterface(Components.interfaces.nsIHttpChannel);
  },

  setURI: function(url) {
    this.__url = url;
    var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    var URI = ioService.newURI(url, null, null);

    this.channel = ioService.newChannelFromURI(URI);
  },

  setRedirectLimitation: function(num) {
    this.httpChannel().redirectionLimit = num;
  },

  asyncOpen: function() {
    this.channel.notificationCallbacks = this;
    this.channel.asyncOpen(this, null);
  },

  setPostData: function(data) {
    var upStream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
    upStream.setData(data, data.length);
    var upChannel = this.channel.QueryInterface(Components.interfaces.nsIUploadChannel);
    upChannel.setUploadStream(upStream, "application/x-www-form-urlencoded", -1);

    this.httpChannel().requestMethod = "POST";
  },

  setRequestHeader: function(header, param) {
    this.httpChannel().setRequestHeader(header, param, true);
  },

  getResponseHeader: function(header) {
    this.httpChannel().getResponseHeader(header);
  },

  setAuthHeader: function(user, pass) {
    this.user = user;
    this.pass = pass;
  },

  abort: function() {
    if (this.timer) {
      this.timer.cancel();
    }
    this.channel.cancel(Components.results.NS_BINDING_ABORTED);
    this.cannnel = null;
  },

  onStartRequest: function(request, context) {
    this.responseText = "";
    try {
      this.status = this.httpChannel().responseStatus;
      this.date   = this.httpChannel().getResponseHeader("Date");
    }
    catch (e) {}
  },

  onDataAvailable: function(request, context, stream, offset, length) {
    var scriptableInputStream = 
      Components.classes["@mozilla.org/scriptableinputstream;1"]
        .createInstance(Components.interfaces.nsIScriptableInputStream);
    scriptableInputStream.init(stream);

    this.responseText += scriptableInputStream.read(length);
  },
  
  onStopRequest: function(request, context, status) {
    if (Components.isSuccessCode(status)) {
      if (this.onload) this.onload(this);
    }
    else if (status != Components.results.NS_BINDING_ABORTED) {
      if (this.onerror) this.onerror(this);
    }
    var observer = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    observer.removeObserver(this, "http-on-modify-request");
    observer.removeObserver(this, "http-on-examine-response");
  },

  onChannelRedirect: function(oldChannel, newChannel, flags) {
    if (this._onredirect) {
      this._onredirect(oldChannel, newChannel, flags);
    }
    else {
      this.channel = newChannel;
    }
  },

  observe: function(subject, topic, data) {
    // Do not use user cookies
    //
    if (subject == this.channel) {
      if (topic == "http-on-modify-request") {
        this.httpChannel().setRequestHeader("Cookie", "", false);
      }
      else if (topic == "http-on-examine-response") {
        this.httpChannel().setResponseHeader("Set-Cookie", "", false);
      }

      if (topic == "http-on-modify-request" && this.user) {
        this.httpChannel().setRequestHeader("Authorization", "Basic " + btoa(toUTF8Octets(this.user + ":" + this.pass)), false);
      }
    }
  },

  // nsIInterfaceRequestor
  getInterface: function(aIID) {
    try {
      return this.QueryInterface(aIID);
    }
    catch (e) {
      throw Components.results.NS_NOINTERFACE;
    }
  },

  // nsIProgressEventSink (to shut up annoying debug exceptions
  onProgress: function(request, context, progress, progressmax) {},
  onStatus: function(request, context, status, statusArg) {},
  
  // nsIHttpEventSink (to shut up annoying debug exceptions
  onRedirect: function(oldChannel, newChannel) {},

  // nsIAuthPromptProvider (to shut up annoying debug exceptions
  getAuthPrompt: function(reason) {},

  QueryInterface: function(aIID) {
    if (aIID.equals(Components.interfaces.nsISupports) ||
        aIID.equals(Components.interfaces.nsIObserver) ||
        aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
        aIID.equals(Components.interfaces.nsIWebProgress) ||
        aIID.equals(Components.interfaces.nsIDocShell) ||
        aIID.equals(Components.interfaces.nsIDocShellTreeItem) ||
        aIID.equals(Components.interfaces.nsIPrompt) ||
        aIID.equals(Components.interfaces.nsIAuthPrompt) ||
        aIID.equals(Components.interfaces.nsIAuthPromptProvider) ||
        aIID.equals(Components.interfaces.nsIInterfaceRequestor) ||
        aIID.equals(Components.interfaces.nsIChannelEventSink) ||
        aIID.equals(Components.interfaces.nsIProgressEventSink) ||
        aIID.equals(Components.interfaces.nsIHttpEventSink) ||
        aIID.equals(Components.interfaces.nsIStreamListener))
      return this;

    throw Components.results.NS_NOINTERFACE;
  }
};

/////////////////////////////////////////////////////////////////////////
//
// Session manager.
//
function Session() {
  this._console = null;
}

Session.prototype = {
  setDelayTask: function(delay, func, data, type) {
    var timer = Components.classes["@mozilla.org/timer;1"] 
      .createInstance(Components.interfaces.nsITimer); 

    var target = this;

    if (type == null) {
      type = Components.interfaces.nsITimer.TYPE_ONE_SHOT;
    }

    timer.initWithCallback({
      notify: function() {
          target[func](data);
        }
      },
      delay,
      type);
    return timer;
  },

  createRequest: function(func, param, isPost) {
    var ns = func.split(".");
    var requestURL = "";
    var requestParam = "";

    var request = new HttpRequest;

    if (ns.length > 1) {
      request.callback = ns[0] + "_" + ns[1];
      requestURL       = ns.join("/");
    }
    else {
      request.callback = func;
      requestURL       = func;
    }

    requestURL = TWITTER_API_URL + requestURL + ".json";

    if (!param) param = {};

    for (var attr in param) {
      if (requestParam) requestParam += "&";
      requestParam += attr + "=" + encodeURIComponent(param[attr]);
    }

    if (isPost) {
      request.setURI(requestURL);
    }
    else {
      request.setURI(requestURL + '?' + requestParam);
    }

    var target = this;
    request.onload  = function(p) { target._onload(p)  };
    request.onerror = function(p) { target._onerror(p) };

    if (isPost) {
      request.setPostData(requestParam);
    }
    this.log(requestURL);

    request.setAuthHeader(this._user, this._pass);
    // set timeout
    request.timer = this.setDelayTask(NETWORK_TIMEOUT_TIME * 1000, "_ontimeout", request);

    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    request.asyncOpen();
    return request;
  },

  post: function(func, params) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }
    this.createRequest(func, params, true);
  },

  get: function(func, params) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }
    this._req = this.createRequest(func, params, false);
  },

  notifyStatus: function(sts, obj) {

    var msg = {"state": sts, "data": obj};

    Components.classes["@mozilla.org/observer-service;1"]
        .getService(Components.interfaces.nsIObserverService)
    .notifyObservers(null, "twitternotifier-status", msg.toSource());
  },

  _onload: function(req) {
    if (req.timer) {
      req.timer.cancel();
    }

    this.log("Response: " + req.status + " (" + req.__url + ")");

    this.onLoad(req);
  },

  _onerror: function(req) {
    this.reportError("Request error occurred in " + req.callback + ": " + req.status);
    if (req.timer) {
      req.timer.cancel();
    }

    if (this[req.callback + "_error"]) {
      this[req.callback + "_error"](req);
    }
    else {
      this.onError(req);
    }

  },

  _ontimeout: function(req) {
    this.reportError("Request timeout occurred in " + req.__url);
    req.abort();

    if (this[req.callback + "_timeout"]) {
      this[req.callback + "_timeout"](req);
    }
    else {
      this.onTimeout(req);
    }

  },

  onLoad:    function(req) {},
  onError:   function(req) {},
  onTimeout: function(req) {},

  reportError: function(msg) {
    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.debug")) {
      Components.utils.reportError(msg);
    }
    this.log(msg);
  },

  log: function(msg) {
    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.debug")) {
      if (this._console == null) 
        this._console = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
      this._console.logStringMessage(msg);
      dump(msg + "\n");
    }
  }
};

/////////////////////////////////////////////////////////////////////////
//
// TwitterNotifier main component.
//
function TwitterNotifier() {
  var obs = Components.classes["@mozilla.org/observer-service;1"]
    .getService(Components.interfaces.nsIObserverService);

  obs.addObserver(this, "xpcom-shutdown", false);
  obs.addObserver(this, "twitternotifier-command", false);

  /*
   * Disable this feature
   *
  try {
    var idleService = Components.classes["@mozilla.org/widget/idleservice;1"]
      .getService(Components.interfaces.nsIIdleService);
    idleService.addIdleObserver(this, 15 * 60); // Now it's fixed. 15 minutes.
  }
  catch (e) {}
  */

  // Setup sound notification
  this._sound = Components.classes["@mozilla.org/sound;1"].createInstance(Components.interfaces.nsISound);

}

// This is the implementation of your component.
var twitternotifier_prototypes = {
  _timer: null,
  _accounts: [],
  _sessions: [],
  _newMessages: [],
  _allMessages: {},
  _tinyURLs: [],
  _idle: STATE_ACTIVE,
  _decodedTinyURLs: [],
  _getDirectMessage: true,

  // for nsISupports
  QueryInterface: function(aIID) {
    // add any other interfaces you support here
    if (!aIID.equals(Components.interfaces.nsISupports) && 
        !aIID.equals(Components.interfaces.nsIObserver) &&
        !aIID.equals(Components.interfaces.nsITwitterNotifier))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  // for nsIObserver
  //
  observe: function(subject, topic, data) { 
    switch (topic) {

    case "twitternotifier-command":
      this.handleCommand(data);
      break;

    case "xpcom-shutdown":
      this.destroy();
      break;

    case "idle":
      this._idle = STATE_IDLE;
      break;

    case "back":
      if (this._idle == STATE_SUSPEND) {
        // set state active before call update tweets
        this._idle = STATE_ACTIVE;
        this.updateTweets();
      }
      this._idle = STATE_ACTIVE;
      break;

    }
  },

  handleCommand: function(data) {
    var msg = eval('(' + data + ')');
    this[msg.command](msg);
  },

  //
  // commands
  //
  initSession: function(user) {
    if (this._user == null) {
      if (this._timer) {
        this._timer.cancel();
      }

      this._user = user.user;
      this._pass = user.pass;

      this.updatePref();
      this.updateTweets();
    }
    else {
      // Send username and turn on twitterfox icon
      this.notifyStatus("updateUsername", this._accounts[this._user].username);
    }
  },

  updatePref: function() {
    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    this._interval = pref.getIntPref("extensions.twitternotifier.interval") * 60 * 1000;
    // fail safe
    if (this._interval < 180 * 1000) {
      this._interval = 180 * 1000;
    }

    if (this._rateLimit) {
      this.setInterval();
    }

    var session = pref.getCharPref("extensions.twitternotifier.session").split(/;/);
    for (var i in session) {
      var ids = session[i].split(/,/);
      if (ids[0]) {
        this._sessions[ids[0]] = {
          timeline: ids[1] || 0,
          messages: ids[2] || 0,
          replies:  ids[3] || 0,
        };
      }
    }
  },

  setNextTimer: function() {
    this._timer = this.setDelayTask(this._interval, "updateTweets", null);
  },

  updateTweets: function() {
    // don't get status.
    if (this._user == null) return;

    if (this._timer) {
      this._timer.cancel();
    }

    this._newMessages = [];

    if (!this._accounts[this._user]) {

      this._accounts[this._user] = {
        timeline: [],
        messages: [],
        replies:  [],
        username: ""
      };

      if (!this._sessions[this._user]) {
        this._sessions[this._user] = {timeline: 0, messages: 0, replies: 0};
      }

      if (this._user.match(/@/) && 
          !this._accounts[this._user].username) {
        this.get("statuses.user_timeline", {"count": 1});
      }
      else {
        this._accounts[this._user].username = this._user;
        this.notifyStatus("updateUsername", this._accounts[this._user].username);
        this.getTweets();
      }

    }
    else {
      this.getTweets();
    }
  },

  getLatestId: function(type) {
    msgs = this._accounts[this._user][type];
    if (msgs.length) {
      var username = this._accounts[this._user].username.toLowerCase();
      for (var i in msgs) {
        var user = msgs[i].user ? msgs[i].user : msgs[i].sender;
        if (user.screen_name.toLowerCase() != username) {
          return {since_id: msgs[i].id, count:200};
        }
      }
    }
    return null;
  },

  getTweets: function() {
    if (this._idle == STATE_ACTIVE) {
      this.get("statuses.friends_timeline", this.getLatestId('timeline'));
    }
    else {
      this._idle = STATE_SUSPEND;
    }
  },

  getReplies: function() {
    this.get("statuses.replies", this.getLatestId('replies'));
  },

  getDirectMessages: function() {
    if (this._getDirectMessage) {
      this.get("direct_messages", this.getLatestId('messages'));
    }
    else {
      this.updateTimeline();
    }
    this._getDirectMessage = !this._getDirectMessage;
  },

  getRecent: function(obj) {
    var type = obj.type;
    if (this._accounts[this._user]) {
      var msg = this.getNumUnread();
      msg['msgs'] = this._accounts[this._user][type];
      msg['type'] = type;
      this.notifyStatus("showPopup", msg);
    }
  },

  getNumUnread: function() {
    if (!this._accounts[this._user]) return 0;

    var result = {};
    var types = ['timeline', 'replies', 'messages'];
    for (var i in types) {
      result[types[i]] = this.countUnread(this._accounts[this._user][types[i]]);
    }
    return result;
  },

  getUnreadCount: function() {
    var ret = this.getNumUnread();
    var unread = 0;
    for (var i in ret) {
      unread += ret[i];
    }
    return unread;
  },

  markRead: function(obj) {
    var type = obj.type;
    for (var i in this._accounts[this._user][type]) {
      this._accounts[this._user][type][i].unread = false;
    }
  },

  markAllRead: function() {
    var types = ['timeline', 'replies', 'messages'];
    for (var i in types) {
      this.markRead({type:types[i]});
    }
  },

  countUnread: function(msgs) {
    var count = 0;
    var username = this._accounts[this._user].username.toLowerCase();
    for (var i in msgs) {
      var user = msgs[i].user ? msgs[i].user : msgs[i].sender;
      try {
        if (msgs[i].unread && user && 
            user.screen_name.toLowerCase() != username) {
          ++count;
        }
      }
      catch (e) {}
    }
    return count;
  },

  sendMessage: function(msg) {
    var arr = /^d\s+(\S+)\s+(.*)/(msg.status);
    if (arr && arr.length == 3) {
      this.post("direct_messages.new", {user: arr[1], text: arr[2]});
    }
    else {
      msg.source = APP_NAME;
      var status = {status:msg.status, source:APP_NAME};
      if (msg.inReplyTo) {
        status["in_reply_to_status_id"] = msg.inReplyTo;
      }
      this.post("statuses.update", status);
    }
  },

  setFavorite: function(msg) {
    this.post("favorites."  + msg.method + "." + msg.id);
  },

  deleteTweet: function(obj) {
    if (obj.type == 'timeline' || obj.type == 'replies') {
      this.post("statuses.destroy", {id:obj.id});
    }
    else {
      this.post("direct_messages.destroy", {id:obj.id});
    }
  },

  changeAccount: function(account) {

    // reset session
    this.logout();
    
    if (!this._accounts[account.user]) {
      this.initSession(account);
    }
    else {
      this._user = account.user;
      this._pass = account.pass;
      this.notifyStatus("updateUsername", this._accounts[this._user].username);
      this.notifyStatus("setUserIcon", this._accounts[this._user].usericon);

      this.updateTweets();
    }
    this.notifyStatus("accountChanged");
  },

  logout: function() {
    if (this._timer) {
      this._timer.cancel();
    }

    if (this._req) {
      this._req.abort();
    }

    this._user = null;
    this._pass = null;
    this._hasError = false;
    this._allMessages = {}
    this._getDirectMessage = true;
  },

  //
  // Private methods.
  //
  destroy: function(e) {
    if (this._timer) {
      this._timer.cancel();
      this._timer = null;
    }
    if (this._req) {
      this._req.abort();
    }

    var obs = Components.classes["@mozilla.org/observer-service;1"]
      .getService(Components.interfaces.nsIObserverService);

    obs.removeObserver(this, "xpcom-shutdown");
    obs.removeObserver(this, "twitternotifier-command");

    try {
      var idleService = Components.classes["@mozilla.org/widget/idleservice;1"]
        .getService(Components.interfaces.nsIIdleService);
      idleService.removeIdleObserver(this, 15 * 60); // Now it's fixed. 15 minutes.
    }
    catch (e) {}

 },

  sortById: function(a, b) {
    return b.id - a.id;
  },

  sortByDate: function(a, b) {
    var ta = new Date(a.created_at);
    var tb = new Date(b.created_at);
    return tb - ta;
  },

  retrieveTimeline: function(obj, req, method) {
    if (obj) {
      this.convertTinyURL(obj);

      var hash = {};

      var stored = this._accounts[this._user][method];
      
      // Avoid duplicate messages.
      for (var i in stored) {
        hash[stored[i].id] = 1;
      }

      // Added 'unread' flag if the message is new in any of stored messages
      this.log("Get " + obj.length + " tweets");
      for (var i in obj) {
        if (typeof obj[i].id != 'undefined' && 
            !hash[obj[i].id]) {

          if (obj[i].id > this._sessions[this._user][method] && 
              !this._allMessages[obj[i].id]) {
            obj[i].unread = true;
            this._newMessages.push(obj[i]);
          }
          stored.unshift(obj[i]);
        }
        try {
          this._allMessages[obj[i].id] = 1;
        }
        catch (e) {}
      }
      stored.sort(this.sortByDate);

      while (stored.length > MAX_STORED_MESSAGES) {
        if (!stored[stored.length - 1].unread) {
          stored.pop();
        }
        else {
          break;
        }
      }
    }
  },

  updateTimeline: function() {
    if (this._newMessages.length) {
      // Play sound if this window is active
      this.playSound();
      this.notifyStatus("updateFriendsTimeline", this._newMessages);
    }
    else {
      if (!this._hasError) {
        this.notifyStatus("noUpdate");
      }
    }
    // reset error flag
    this._hasError = false;

    this.decodeTinyURL();
    this.storeSession();

    this._newMessages = [];
    this.setNextTimer();
  },

  playSound: function() {
    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.sound")) {

      try {
        var IOService = Components.classes['@mozilla.org/network/io-service;1'].getService(Components.interfaces.nsIIOService);
        var localFile = Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile);
        var url = pref.getCharPref("extensions.twitternotifier.soundFile");
      
        localFile.initWithPath(url);

        this._sound.play(IOService.newFileURI(localFile));
      }
      catch (e) {
      }
    }
  },

  convertTinyURL: function(msg) {
    var re = /http:\/\/(tinyurl.com|is.gd|bit.ly)\/\w+/;

    for (var i in msg) {

      var text = msg[i].text;
      while(re.exec(text)) {
        var url = RegExp.lastMatch;
        text = RegExp.rightContext;
        this._tinyURLs.push(url);
      }
    }
  },

  decodeTinyURL: function() {

    while (1) {

      if (this._tinyURLs.length == 0) {
        return;
      }
      var url = this._tinyURLs[0];

      if (!this._decodedTinyURLs[url]) {
        break;
      }
      this._tinyURLs.shift();
    }

    var target = this;
    var req = new HttpRequest;
    req.setURI(url);
    req.setRedirectLimitation(0);
    req.onload     = function(e) {target.onDecodeTinyURL(e);}
    req.onerror    = function(e) {target.onErrorDecodeTinyURL(e);};
    req._onredirect = function(oldChannel, newChannel, flags) {target.storeTinyURL(oldChannel, newChannel, url); }
    req.asyncOpen(true);
  },

  storeTinyURL: function(oldChannel, newChannel, url) {
    var newurl = newChannel.URI.prePath + newChannel.URI.path;
    this._decodedTinyURLs[url] = newurl;

    this._tinyURLs.shift();
    this.decodeTinyURL();
  },

  storeSession: function() {

    var account = this._accounts[this._user];

    var types = ['timeline', 'replies', 'messages'];
    for (var i in types) {
      if (account[types[i]].length) {
        this._sessions[this._user][types[i]] = account[types[i]][0].id;
      }
    }

    var storage = [];

    for (var i in this._sessions) {
      storage.push([i, 
                    this._sessions[i].timeline,
                    this._sessions[i].messages, 
                    this._sessions[i].replies].join(','));
    }
    var pref = Components.classes['@mozilla.org/preferences-service;1']
      .getService(Components.interfaces.nsIPrefBranch);

    pref.setCharPref("extensions.twitternotifier.session", storage.join(';'));
  },

  getDecodedTinyURL: function(url) {
    return this._decodedTinyURLs[url];
  },

  onDecodeTinyURL: function(e) {
    this._tinyURLs.shift();
    this.decodeTinyURL();
  },

  onErrorDecodeTinyURL: function(req) { 
    this._decodedTinyURLs[req.__url] = req.httpChannel().getResponseHeader("Location");
    this._tinyURLs.shift();
    this.setDelayTask(1 * 1000, "decodeTinyURL", null);
  },

  deleteMessage: function(obj, type) {
    var msgs = this._accounts[this._user][type];
    for (var i = 0; i < msgs.length; ++i) {
      if (msgs[i].id == obj.id) {
        this._accounts[this._user][type].splice(i, 1);
        break;
      }
    }
  },

  // Network handler
  //
  onLoad: function(req) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }

    switch (Number(req.status)) {
    case 400:
      this.rateLimitExceeded(req);
      break;

    case 401:
      if (req.callback == "statuses_friends_timeline" ||
          !this.dispatchError(req)) {
        this._accounts[this._user] = null;
        this._user = null;
        this._pass = null;
        this.notifyStatus("authFail");
      }
      break;

    case 403:
    case 404:
    case 500:
    case 502:
    case 503:
      this.handleError(req, "Twitter server responded with an error (" + req.status + ")");
      break;
 
    case 200:
    case 304:
    default:
      var resp = null;
      if (!req.responseText.match(/^\s*$/)) {
        req.responseText = req.responseText.replace(/Couldn\'t find Status with ID=\d+,/, '');
        try {
          var resp = JSON.fromString(req.responseText);
        }
        catch (e) {
          this.reportError("An error occurred while requesting " + req.__url);
          this.log("Response text: " + e.message);
          this.handleError(req, "Can't parse JSON. Twitter server responded with an error.");
        }
      }

      if (resp == null || (resp && resp.error)) {
        if (resp && resp.error) {
          this.notifyStatus("internalError", resp.error);
          this.setNextTimer();
        }
        else {
          this.handleError(req, "Twitter server responded with an error");
        }
      }
      else {
        this[req.callback](resp.statuses ? resp.statuses : resp, req);
      }
      break;
    }
  },

  handleError: function(req, msg) {
    if (this[req.callback + "_error"]) {
      this[req.callback + "_error"](req);
    }
    else {
      if (!this.dispatchError(req)) {
        this.notifyStatus("internalError", msg);
        this.setNextTimer();
      }
    }
  },

  dispatchError: function(req) {

    this._hasError = true;

    switch (req.callback) {
    case "direct_messages":
      this.updateTimeline();
      break;
    
    case "statuses_replies":
      this.getDirectMessages();
      break;
    
    case "statuses_friends_timeline":
      this.getReplies();
      break;
    
    default:
      return false;
    }
    return true;
  },

  onError: function(req) {

    this.notifyStatus("internalError", "Network error with " + req.callback);
    if (this.dispatchError(req)) return;

    if (this._timer) {
      this._timer.cancel();
    }
    this.setNextTimer();
  },

  onTimeout: function (req) {
    this.notifyStatus("internalError", "Request timeout with " + req.callback);
    if (this.dispatchError(req)) return;

    if (this._timer) {
      this._timer.cancel();
    }
    this.setNextTimer();
  },

  rateLimitExceeded: function(req) {
    try {
      var resp = JSON.fromString(req.responseText);
      this.notifyStatus("internalError", resp.error);
    }
    catch (e) {}

    this.get('account.rate_limit_status');
  },

  setInterval: function() {
    var interval = this._rateLimit / 3;
    interval = Math.ceil((60 / interval) * 60 * 1000);
    
    if (this._interval < interval) {
      this._interval = interval;
    }

    // fail safe
    if (this._interval < 180 * 1000) {
      this._interval = 180 * 1000;
    }
  },

  //
  // TwitterAPI callbacks.
  //
  account_rate_limit_status: function(obj, req) {
    this._rateLimit = obj.hourly_limit;
    this.log("Rate limit is " + this._rateLimit + " requests per hour");
    if (this._rateLimit <= 3) {
      this._rateLimit = 3;
    }

    this.setInterval();

    var nextTime = new Date(obj.reset_time) - Date.now();
    if (nextTime < 180 * 1000) {
      nextTime = 180 * 1000;
    }
    else if (nextTime > 3600 * 1000) {
      nextTime = 3600 * 1000;
    }
    this._timer = this.setDelayTask(nextTime, "updateTweets", null);

    var msg = "Twitter lower rate limit (less than " + this._rateLimit + 
              " requests per hour), request interval has been changed to " + this._interval / 60 / 1000 + " minutes"; 
    this.notifyStatus("showMessage", msg);
    this.log(msg);
    this.log("Auto rate limit control:" + obj.remaining_hits + "/" + obj.hourly_limit + 
             ", reset time: " + obj.reset_time + "(" + Math.ceil(nextTime / 1000 / 60) + " minutes)");
  },

  direct_messages: function(obj, req) {
    this.retrieveTimeline(obj, req, "messages");

    try {
      this._accounts[this._user].usericon = obj[0].recipient.profile_image_url;
      this.notifyStatus("setUserIcon", obj[0].recipient.profile_image_url);
    }
    catch (e) {}
    

    this.updateTimeline();
  },

  statuses_replies: function(obj, req) {
    this.retrieveTimeline(obj, req, "replies");
    this.getDirectMessages();
  },

  statuses_friends_timeline: function(obj, req) {
    this.retrieveTimeline(obj, req, "timeline");
    this.getReplies();
  },

  statuses_user_timeline: function(obj) {
    try {
      this._accounts[this._user].username = obj[0].user.screen_name;
      this.notifyStatus("updateUsername", obj[0].user.screen_name);
      this.updateTweets();
    }
    catch (e) {
      this.setNextTimer();
    }
  },

  statuses_destroy: function(obj) {
    this.notifyStatus("messageDeleted", {id:obj.id});
    this.deleteMessage(obj, 'timeline');
    this.deleteMessage(obj, 'replies');
  },

  favorites_destroy: function(obj) {
    this.toggleFavorite(obj.id, null);
    this.notifyStatus("updateFavorite", {id: obj.id, state:null});
  },

  favorites_create: function(obj) {
    this.toggleFavorite(obj.id, true);
    this.notifyStatus("updateFavorite", {id: obj.id, state:true});
  },

  toggleFavorite: function(id, flag) {
    var types = ['timeline', 'replies'];
    for (var i in types) {
      var msgs = this._accounts[this._user][types[i]];
      for (var j in msgs) {
        if (msgs[j].id == id) {
          msgs[j].favorited = flag;
          break;
        }
      }
    }
  },

  statuses_update: function(obj, req) {
    if (obj.id) {
      this._accounts[this._user].timeline.unshift(obj);
      this.notifyStatus("sentMessage", obj);
    }
    else {
      this.notifyStatus("errorOnSendMessage");
    }
  },

  statuses_update_error: function(req) {
    this.reportError("Send error occurred: " + req.status);
    this.log(req.responseText);
    this.notifyStatus("errorOnSendMessage");
  },

  statuses_update_timeout: function() {
    this.reportError("Send timeout occurred");
    this.notifyStatus("errorOnSendMessage");
  },

  direct_messages_new: function(obj, req) {
    if (obj.id) {
      this.notifyStatus("sentMessage", obj);
    }
    else {
      this.notifyStatus("errorOnSendMessage");
    }
  },

  direct_messages_destroy: function(obj) {
    this.notifyStatus("messageDeleted", {id:obj.id});
    this.deleteMessage(obj, 'messages');
  },

  direct_messages_new_error: function(req) {
    this.reportError("Send error occurred: " + req.status);
    this.log(req.responseText);
    this.notifyStatus("errorOnSendMessage");
  },

  direct_messages_new_timeout: function() {
    this.reportError("Send timeout occurred");
    this.notifyStatus("errorOnSendMessage");
  },
}

TwitterNotifier.prototype = new Session;

for (var i in twitternotifier_prototypes) {
  TwitterNotifier.prototype[i] = twitternotifier_prototypes[i];
}

//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//
// Singleton
var gTwitterNotifier = null;

// Factory
var TwitterNotifierFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    if (gTwitterNotifier === null) {
      gTwitterNotifier = new TwitterNotifier().QueryInterface(aIID);
    }
    return gTwitterNotifier;
  }
};

// Module
var TwitterNotifierModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);

    Components.classes["@mozilla.org/categorymanager;1"]
      .getService(Components.interfaces.nsICategoryManager)
        .addCategoryEntry("app-startup", 
                          CLASS_NAME,
                          "service," + CONTRACT_ID,
                          true, true);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        

    Components.classes["@mozilla.org/categorymanager;1"]
      .getService(Components.interfaces.nsICategoryManager)
        .deleteCategoryEntry("app-startup", 
                             CLASS_NAME,
                             true);
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return TwitterNotifierFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

//module initialization
function NSGetModule(aCompMgr, aFileSpec) { return TwitterNotifierModule; }

